import logging
import os

from flask import Flask

# Lendo variáveis de ambiente
APP_NAME = os.getenv("APP_NAME", "nsj_rest_lib")
MOPE_CODE = os.getenv("MOPE_CODE")
DEFAULT_PAGE_SIZE = int(os.getenv("DEFAULT_PAGE_SIZE", 20))
USE_SQL_RETURNING_CLAUSE = (
    os.getenv("USE_SQL_RETURNING_CLAUSE", "true").lower() == "true"
)

DATABASE_HOST = os.getenv("DATABASE_HOST", "")
DATABASE_PASS = os.getenv("DATABASE_PASS", "")
DATABASE_PORT = os.getenv("DATABASE_PORT", "")
DATABASE_NAME = os.getenv("DATABASE_NAME", "")
DATABASE_USER = os.getenv("DATABASE_USER", "")
DATABASE_DRIVER = os.getenv("DATABASE_DRIVER", "POSTGRES")

DB_POOL_SIZE = int(os.getenv("DB_POOL_SIZE", 1))

CLOUD_SQL_CONN_NAME = os.getenv("CLOUD_SQL_CONN_NAME", "")
ENV = os.getenv("ENV", "")


def get_logger():
    return logging.getLogger(APP_NAME)


application = Flask("app")



import abc
import copy
import enum

# import uuid

from typing import Any, Dict, List, Set, Union

from nsj_rest_lib.entity.entity_base import EMPTY, EntityBase
from nsj_rest_lib.descriptor.conjunto_type import ConjuntoType
from nsj_rest_lib.descriptor.dto_field import DTOField, DTOFieldFilter


class DTOBase(abc.ABC):
    resume_fields: Set[str] = set()
    partition_fields: Set[str] = set()
    fields_map: Dict[str, DTOField] = {}
    list_fields_map: dict = {}
    left_join_fields_map: dict = {}
    left_join_fields_map_to_query: dict = {}
    sql_join_fields_map: dict = {}
    sql_join_fields_map_to_query: dict = {}
    sql_read_only_fields: list = []
    object_fields_map: list = []
    field_filters_map: Dict[str, DTOFieldFilter]
    # TODO Refatorar para suportar PK composto
    pk_field: str
    fixed_filters: Dict[str, Any]
    conjunto_type: ConjuntoType
    conjunto_field: str
    escape_validator: bool
    uniques: Dict[str, Set[str]]
    candidate_keys: List[str]
    search_fields: Set[str]

    def __init__(
        self,
        entity: Union[EntityBase, dict] = None,
        escape_validator: bool = False,
        generate_default_pk_value: bool = True,
        validate_read_only: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()

        self.escape_validator = escape_validator
        self.generate_default_pk_value = generate_default_pk_value

        # Transformando a entity em dict (se houver uma entity)
        if entity is not None:
            kwargs = (
                copy.deepcopy(entity)
                if type(entity) is dict
                else copy.deepcopy(entity.__dict__)
            )

        # Setando os campos registrados como fields simples
        for field in self.__class__.fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.fields_map[field]

            if (
                validate_read_only
                and aux_dto_field.read_only
                and kwargs.get(field, None) is not None
            ):
                raise ValueError(f"O campo {field} não pode ser preenchido.")

            # Tratando do valor default
            if (
                aux_dto_field.default_value is not None
                and kwargs.get(field, None) is None
                and (not aux_dto_field.pk or generate_default_pk_value)
            ):
                default_value = aux_dto_field.default_value
                if callable(aux_dto_field.default_value):
                    default_value = aux_dto_field.default_value()
                kwargs[field] = default_value

            # Verificando se é preciso converter o nome do field para o nome correspondente no Entity
            # E, se será preciso aplicar alguma conversão customizada (para trazer o valor do entity para o DTO)
            entity_field = field
            if entity is not None:
                if aux_dto_field.entity_field is not None:
                    entity_field = aux_dto_field.entity_field

                # Verificando se o campo carece de conversão customizada
                if aux_dto_field.convert_from_entity is not None:
                    fields_converted = aux_dto_field.convert_from_entity(
                        kwargs[entity_field], kwargs
                    )
                    if field not in fields_converted:
                        setattr(self, field, None)

                    for converted_key in fields_converted:
                        setattr(self, converted_key, fields_converted[converted_key])

                    continue

            # Atribuindo o valor à propriedade do DTO
            if entity_field in kwargs:
                setattr(self, field, kwargs[entity_field])
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields de join por meio da consulta SQL
        for field in self.__class__.sql_join_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.sql_join_fields_map[field]

            # Verificando a conversão do nome do field para o nome correspondente no Entity (no DTO Relacionado)
            # E, se será preciso aplicar alguma conversão customizada (para trazer o valor do entity para o DTO)
            entity_field = field
            if entity is not None:
                if (
                    not aux_dto_field.related_dto_field
                    in aux_dto_field.dto_type.fields_map
                ):
                    continue

                entity_field = aux_dto_field.dto_type.fields_map[
                    aux_dto_field.related_dto_field
                ].get_entity_field_name()

                # Verificando se o campo carece de conversão customizada
                if aux_dto_field.convert_from_entity is not None:
                    fields_converted = aux_dto_field.convert_from_entity(
                        kwargs[entity_field], kwargs
                    )
                    if field not in fields_converted:
                        setattr(self, field, None)

                    for converted_key in fields_converted:
                        setattr(self, converted_key, fields_converted[converted_key])

                    continue

            # Atribuindo o valor à propriedade do DTO
            if entity_field in kwargs:
                setattr(self, field, kwargs[entity_field])
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields left join
        for field in self.__class__.left_join_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.left_join_fields_map[field]

            # Atribuindo o valor à propriedade do DTO
            if field in kwargs:
                setattr(self, field, kwargs[field])
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields object
        for field in self.__class__.object_fields_map:
            # Recuperando a configuração do campo
            aux_dto_field = self.__class__.object_fields_map[field]

            # Atribuindo o valor à propriedade do DTO
            if field in kwargs:
                setattr(self, field, aux_dto_field.expected_type(**kwargs[field]))
            else:
                setattr(self, field, None)

        # Setando os campos registrados como fields de lista
        if entity is None:
            for field in self.__class__.list_fields_map:
                dto_list_field = self.__class__.list_fields_map[field]

                if field in kwargs:
                    if not isinstance(kwargs[field], list):
                        raise ValueError(
                            f"O campo {field} deveria ser uma lista do tipo {dto_list_field.dto_type}."
                        )

                    related_itens = []
                    for item in kwargs[field]:
                        # Preenchendo os campos de particionanmento, se necessário (normalmente: tenant e grupo_empresarial)
                        for partition_field in self.__class__.partition_fields:
                            if (
                                (
                                    partition_field not in item
                                    or item[partition_field] is None
                                )
                                and partition_field
                                in dto_list_field.dto_type.partition_fields
                            ):
                                partition_value = getattr(self, partition_field)
                                item[partition_field] = partition_value

                        # Criando o DTO relacionado
                        item_dto = dto_list_field.dto_type(**item)

                        # Adicionando o DTO na lista do relacionamento
                        related_itens.append(item_dto)

                    setattr(self, field, related_itens)
                else:
                    setattr(self, field, None)

        # Tratando do ID automático
        # if generate_pk_uuid:
        #     if getattr(self, self.__class__.pk_field) is None:
        #         setattr(self, self.__class__.pk_field, uuid.uuid4())

    def convert_to_entity(self, entity_class: EntityBase, none_as_empty: bool = False):
        """
        Cria uma instância da entidade, e a popula com os dados do DTO
        corrente.

        É importante notar que as equivalências dos nomes dos campos
        são tratadas neste método.
        """

        entity = entity_class()

        for field, dto_field in self.__class__.fields_map.items():
            # Verificando se é preciso realizar uma tradução de nome do campo
            entity_field = field
            if dto_field.entity_field is not None:
                entity_field = dto_field.entity_field

            # Verificando se o campo existe na entity
            if not hasattr(entity, entity_field):
                continue

            # Recuperando o valor
            value = getattr(self, field)

            # Verificando se é necessária alguma conversão customizada
            custom_value_conveted = DTOBase.custom_convert_value_to_entity(
                value,
                dto_field,
                entity_field,
                none_as_empty,
                self.__dict__,
            )
            for key in custom_value_conveted:
                setattr(entity, key, custom_value_conveted[key])

            if len(custom_value_conveted) > 0:
                continue

            # Convertendo o value para o correspondente nos fields
            value = DTOBase.convert_value_to_entity(
                value,
                dto_field,
                none_as_empty,
                entity_class,
            )

            # Setando na entidade
            setattr(entity, entity_field, value)

        return entity

    def custom_convert_value_to_entity(
        value: Any,
        dto_field: DTOField,
        entity_field: str,
        none_as_empty: bool,
        dto_values: Dict[str, any],
    ):
        # Verificando se é necessária uma conversão customizada
        retorno = {}
        if value is not None and dto_field.convert_to_entity is not None:
            fields_converted = dto_field.convert_to_entity(value, dto_values)
            if entity_field not in fields_converted:
                retorno[entity_field] = None if not none_as_empty else EMPTY

            for converted_key in fields_converted:
                value = fields_converted[converted_key]
                if value is None and none_as_empty:
                    value = EMPTY
                retorno[converted_key] = value

        return retorno

    def convert_value_to_entity(
        value: Any,
        dto_field: DTOField,
        none_as_empty: bool,
        entity_class: EntityBase,
    ) -> Any:
        # Enumerados
        if isinstance(dto_field.expected_type, enum.EnumMeta):
            try:
                # TODO Rever se não é para verificar se o valor já é do tipo enumerado, e só retornar seu value
                # (ou índice da tupla dentro do value, se houver uma tupla)
                # Porque parece que ele tenta sempre converter para o tipo esperado, mas isso pode já ter sido
                # feito no DTO
                return DTOBase._convert_enum_to_entity(value, dto_field)
            except ValueError:
                # Retornando o pórpio valor, caso se deseje converter um enumerado que não seja válido
                # Isso é, aceitando enumerados inválidos (só para os filtros)
                return value.value if isinstance(value, enum.Enum) else str(value)

        # Bool to int
        entity_fields_map = getattr(entity_class, "fields_map", {})
        entity_field_name = dto_field.entity_field or dto_field.name
        if (
            isinstance(value, bool)
            and entity_field_name in entity_fields_map
            and entity_fields_map[entity_field_name].expected_type == int
        ):
            return 1 if value else 0

        # Convertendo None para EMPTY (se desejado)
        if value is None:
            if none_as_empty:
                return EMPTY
            else:
                return value

        return value

    def _convert_enum_to_entity(
        value: Any,
        dto_field: DTOField,
    ):
        # Ignorando valores nulos
        if value is None:
            return None

        # Construindo uma lista com os itens do enumerado esperado
        lista_enum = list(dto_field.expected_type)

        # Se o enum estiver vazio
        if len(lista_enum) <= 0:
            return None

        # Verificando o tipo dos valores do enum
        if isinstance(lista_enum[0].value, tuple):
            # Recuperando o item correspondente do enumerado
            enumerado = value
            parou = False
            if not isinstance(value, dto_field.expected_type):
                for item in dto_field.expected_type:
                    lista_valores = list(item.value)
                    for valor in lista_valores:
                        if valor == value:
                            enumerado = item
                            parou = True
                            break

                        # Se o valor for string, testa inclusive em caixa alta e baixa
                        if isinstance(value, str):
                            if valor == value.lower() or valor == value.upper():
                                enumerado = item
                                parou = True
                                break

                    if parou:
                        break

            # Convertendo do enumerado para o valor desejado na entidade
            if "get_entity_index" in enumerado.__dict__:
                tuple_index = enumerado.get_entity_index()
            else:
                tuple_index = 1

            if isinstance(enumerado.value, list) or isinstance(enumerado.value, tuple):
                return enumerado.value[tuple_index]
            else:
                return enumerado.value
        else:
            # Tentando pelo valor do próprio enum (e testando os casos, se for str)
            if isinstance(value, str):
                try:
                    return dto_field.expected_type(value)
                except ValueError:
                    try:
                        return dto_field.expected_type(value.lower())
                    except ValueError:
                        return dto_field.expected_type(value.upper())
            else:
                return dto_field.expected_type(value)

    def convert_to_dict(
        self, fields: Dict[str, List[str]] = None, just_resume: bool = False
    ):
        """
        Converte DTO para dict
        """

        # Resolving fields to use
        if fields is None:
            fields = {"root": self.resume_fields}

        if just_resume:
            fields = {"root": self.resume_fields}
        else:
            fields["root"] = fields["root"].union(self.resume_fields)

        # Making result maps
        result = {}

        # Converting simple fields
        for field in self.fields_map:
            if not field in fields["root"]:
                continue

            result[field] = getattr(self, field)

        # Converting sql join fields
        for field in self.sql_join_fields_map:
            if not field in fields["root"]:
                continue

            result[field] = getattr(self, field)

        # Converting left join fields
        for field in self.left_join_fields_map:
            if not field in fields["root"]:
                continue

            result[field] = getattr(self, field)

        # Converting list fields
        for field in self.list_fields_map:
            if not field in fields["root"]:
                continue

            # Criando o mapa de fileds para a entidade aninhada
            internal_fields = None
            if field in fields:
                internal_fields = {"root": fields[field]}

            # Recuperando o valor do atributo
            value = getattr(self, field)
            if value is None:
                value = []

            # Convetendo a lista de DTOs aninhados
            result[field] = [
                item.convert_to_dict(internal_fields, just_resume) for item in value
            ]

        return result

    def get_entity_field_name(self, dto_field_name: str) -> str:
        """
        Retorna o nome correspondente do field no entity
        (o qual é o nome do field no DTO por padrão, ou o nome que for
        passado no parâmetro "entity_field" no construtor).

        Retorna None, se o "dto_field_name" não for achado como um dos
        fields registrados.
        """

        if dto_field_name not in self.fields_map:
            return None

        return self.fields_map[dto_field_name].get_entity_field_name()



import functools
from typing import Any, Dict

from nsj_rest_lib.descriptor.conjunto_type import ConjuntoType
from nsj_rest_lib.descriptor.dto_field import DTOField
from nsj_rest_lib.descriptor.dto_list_field import DTOListField
from nsj_rest_lib.descriptor.dto_left_join_field import DTOLeftJoinField, LeftJoinQuery
from nsj_rest_lib.descriptor.dto_object_field import DTOObjectField
from nsj_rest_lib.descriptor.dto_sql_join_field import DTOSQLJoinField, SQLJoinQuery


def DTO(
    fixed_filters: Dict[str, Any] = None,
    conjunto_type: ConjuntoType = None,
    conjunto_field: str = None,
    filter_aliases: Dict[str, Any] = None,
):

    def DTOWrapper(classe):

        class DTOClass:

            def __init__(
                self,
            ) -> None:
                super().__init__()

                self._fixed_filters = fixed_filters
                self._conjunto_type = conjunto_type
                self._conjunto_field = conjunto_field
                self._filter_aliases = filter_aliases

                if (
                    self._conjunto_type is None and self._conjunto_field is not None
                ) or (self._conjunto_type is not None and self._conjunto_field is None):
                    raise Exception(
                        "Os parâmetros conjunto_type e conjunto_field devem ser preenchidos juntos (se um for não nulo, ambos devem ser preenchidos)."
                    )

            @functools.wraps(classe)
            def __call__(self, cls: object):
                """
                Iterating DTO class to handle DTOFields descriptors.
                """

                # Creating resume_fields in cls, if needed
                self._check_class_attribute(cls, "resume_fields", set())

                # Creating fields_map in cls, if needed
                self._check_class_attribute(cls, "fields_map", {})

                # Creating list_fields_map in cls, if needed
                self._check_class_attribute(cls, "list_fields_map", {})

                # Creating left_join_fields_map in cls, if needed
                self._check_class_attribute(cls, "left_join_fields_map", {})

                # Creating left_join_fields_map_to_query in cls, if needed
                self._check_class_attribute(cls, "left_join_fields_map_to_query", {})

                # Creating sql_join_fields_map in cls, if needed
                self._check_class_attribute(cls, "sql_join_fields_map", {})

                # Creating sql_join_fields_map_to_query in cls, if needed
                self._check_class_attribute(cls, "sql_join_fields_map_to_query", {})

                # Creating sql_read_only_fields in cls, if needed
                self._check_class_attribute(cls, "sql_read_only_fields", [])

                # Creating object_fields_map in cls, if needed
                self._check_class_attribute(cls, "object_fields_map", {})

                # Creating field_filters_map in cls, if needed
                self._check_class_attribute(cls, "field_filters_map", {})

                # Creating pk_field in cls, if needed
                # TODO Refatorar para suportar PKs compostas
                self._check_class_attribute(cls, "pk_field", None)

                # Criando a propriedade "partition_fields" na classe "cls", se necessário
                self._check_class_attribute(cls, "partition_fields", set())

                # Criando a propriedade "uniques" na classe "cls", se necessário
                self._check_class_attribute(cls, "uniques", {})

                # Criando a propriedade "candidate_keys" na classe "cls", se necessário
                self._check_class_attribute(cls, "candidate_keys", [])

                # Criando a propriedade "search_fields" na classe "cls", se necessário
                self._check_class_attribute(cls, "search_fields", set())

                # Iterating for the class attributes
                for key, attr in cls.__dict__.items():
                    # Test if the attribute uses the DTOFiel descriptor
                    if isinstance(attr, DTOField):
                        # Storing field in fields_map
                        getattr(cls, "fields_map")[key] = attr

                        # Setting a better name to storage_name
                        attr.storage_name = f"{key}"
                        attr.name = f"{key}"

                        # Checking filters name
                        self._check_filters(cls, key, attr)

                        # Copying type from annotation (if exists)
                        if key in cls.__annotations__:
                            attr.expected_type = cls.__annotations__[key]

                        # Checking if it is a resume field (to store)
                        if attr.resume:
                            resume_fields = getattr(cls, "resume_fields")
                            if key not in resume_fields:
                                resume_fields.add(key)

                        # TODO Refatorar para suportar PKs compostas
                        # Setting PK info
                        if attr.pk:
                            setattr(cls, "pk_field", f"{key}")

                        # Verifica se é um campo de particionamento, e o guarda em caso positivo
                        if attr.partition_data:
                            partition_fields = getattr(cls, "partition_fields")
                            if key not in partition_fields:
                                partition_fields.add(key)

                        # Verifica se é um campo pertencente a uma unique, a populando o dicionário de uniques
                        if attr.unique:
                            uniques = getattr(cls, "uniques")
                            fields_unique = uniques.setdefault(attr.unique, set())
                            fields_unique.add(key)

                        # Verifica se é uma chave candidata
                        if attr.candidate_key:
                            getattr(cls, "candidate_keys").append(key)

                        # Verifica se é um campo passível de busca
                        if attr.search:
                            getattr(cls, "search_fields").add(key)

                        # Verifica se um campo é somente para leitura
                        if attr.read_only:
                            getattr(cls, "sql_read_only_fields").append(
                                attr.entity_field or key
                            )

                    elif isinstance(attr, DTOListField):
                        # Storing field in fields_map
                        getattr(cls, "list_fields_map")[key] = attr

                        # Setting a better name to storage_name
                        attr.storage_name = f"{key}"
                        attr.name = f"{key}"

                    elif isinstance(attr, DTOLeftJoinField):
                        # Storing field in fields_map
                        getattr(cls, "left_join_fields_map")[key] = attr

                        # Setting a better name to storage_name
                        attr.storage_name = f"{key}"
                        attr.name = f"{key}"

                        # Copying type from annotation (if exists)
                        if key in cls.__annotations__:
                            attr.expected_type = cls.__annotations__[key]

                        # Checking if it is a resume field (to store)
                        if attr.resume:
                            resume_fields = getattr(cls, "resume_fields")
                            if key not in resume_fields:
                                resume_fields.add(key)

                        # Montando o mapa de controle das queries (para o service_base)
                        self.set_left_join_fields_map_to_query(key, attr, cls)

                    elif isinstance(attr, DTOSQLJoinField):
                        # Storing field in fields_map
                        getattr(cls, "sql_join_fields_map")[key] = attr

                        # Setting a better name to storage_name
                        attr.storage_name = f"{key}"
                        attr.name = f"{key}"

                        # Copying type from annotation (if exists)
                        if key in cls.__annotations__:
                            attr.expected_type = cls.__annotations__[key]

                        # Checking if it is a resume field (to store)
                        if attr.resume:
                            resume_fields = getattr(cls, "resume_fields")
                            if key not in resume_fields:
                                resume_fields.add(key)

                        # Montando o mapa de controle das queries (para o service_base)
                        self.set_sql_join_fields_map_to_query(key, attr, cls)

                    elif isinstance(attr, DTOObjectField):
                        # Storing field in fields_map
                        getattr(cls, "object_fields_map")[key] = attr

                        # Setting a better name to storage_name
                        attr.storage_name = f"{key}"
                        attr.name = f"{key}"

                        # Copying type from annotation (if exists)
                        if key in cls.__annotations__:
                            attr.expected_type = cls.__annotations__[key]

                        # Checking if it is a resume field (to store)
                        if attr.resume:
                            resume_fields = getattr(cls, "resume_fields")
                            if key not in resume_fields:
                                resume_fields.add(key)

                # Setting fixed filters
                setattr(cls, "fixed_filters", self._fixed_filters)

                # Setting tipo de Conjunto
                setattr(cls, "conjunto_type", self._conjunto_type)
                setattr(cls, "conjunto_field", self._conjunto_field)

                # Setting filter aliases
                setattr(cls, "filter_aliases", self._filter_aliases)

                return cls

            def _check_filters(self, cls: object, field_name: str, dto_field: DTOField):
                """
                Check filters (if exists), and setting default filter name.
                """

                if dto_field.filters is None:
                    return

                # Handling each filter
                for filter in dto_field.filters:
                    # Resolving filter name
                    filter_name = field_name
                    if filter.name is not None:
                        filter_name = filter.name

                    # Storing field filter name
                    filter.field_name = field_name

                    # Adding into field filters map
                    field_filters_map = getattr(cls, "field_filters_map")
                    field_filters_map[filter_name] = filter

            def _check_class_attribute(
                self, cls: object, attr_name: str, default_value: Any
            ):
                """
                Add attribute "attr_name" in class "cls", if not exists.
                """

                if attr_name not in cls.__dict__:
                    setattr(cls, attr_name, default_value)

            def set_left_join_fields_map_to_query(
                self,
                field: str,
                attr: DTOLeftJoinField,
                cls: object,
            ):
                # Recuperando o map de facilitação das queries
                left_join_fields_map_to_query: dict[str, LeftJoinQuery] = getattr(
                    cls, "left_join_fields_map_to_query"
                )

                # Verificando se o objeto de query, relativo a esse campo,
                # já estava no mapa (e colocando, caso negativo)
                map_key = f"{attr.dto_type}____{attr.entity_type}____{attr.entity_relation_owner}"
                left_join_query = left_join_fields_map_to_query.setdefault(
                    map_key, LeftJoinQuery()
                )

                # Preenchendo as propriedades que serão úteis para as queries
                left_join_query.related_dto = attr.dto_type
                left_join_query.related_entity = attr.entity_type
                left_join_query.fields.append(field)
                left_join_query.left_join_fields.append(attr)
                left_join_query.entity_relation_owner = attr.entity_relation_owner

            def set_sql_join_fields_map_to_query(
                self,
                field: str,
                attr: DTOSQLJoinField,
                cls: object,
            ):
                # Recuperando o map de facilitação das queries
                sql_join_fields_map_to_query: dict[str, SQLJoinQuery] = getattr(
                    cls, "sql_join_fields_map_to_query"
                )

                # Verificando se o objeto de query, relativo a esse campo,
                # já estava no mapa (e colocando, caso negativo)
                map_key = f"{attr.dto_type}____{attr.entity_type}____{attr.entity_relation_owner}____{attr.join_type}"
                sql_join_query = sql_join_fields_map_to_query.setdefault(
                    map_key, SQLJoinQuery()
                )

                # Preenchendo as propriedades que serão úteis para as queries
                sql_join_query.related_dto = attr.dto_type
                sql_join_query.related_entity = attr.entity_type
                sql_join_query.fields.append(field)
                sql_join_query.related_fields.append(attr.related_dto_field)
                sql_join_query.join_fields.append(attr)
                sql_join_query.entity_relation_owner = attr.entity_relation_owner
                sql_join_query.join_type = attr.join_type
                sql_join_query.relation_field = attr.relation_field

                if sql_join_query.sql_alias is None:
                    sql_join_query.sql_alias = (
                        f"join_table_{len(sql_join_fields_map_to_query)}"
                    )

        return DTOClass().__call__(classe)

    return DTOWrapper



import typing

from decimal import Decimal
from typing import Any

from nsj_rest_lib.descriptor.filter_operator import FilterOperator
from nsj_rest_lib.util.type_validator_util import TypeValidatorUtil


class DTOFieldFilter:
    def __init__(
        self, name: str = None, operator: FilterOperator = FilterOperator.EQUALS
    ):
        self.name = name
        self.operator = operator
        self.field_name = None

    def set_field_name(self, field_name: str):
        self.field_name = field_name


class DTOField:
    _ref_counter = 0

    def __init__(
        self,
        type: object = None,
        not_null: bool = False,
        resume: bool = False,
        min: int = None,
        max: int = None,
        validator: typing.Callable = None,
        strip: bool = False,
        entity_field: str = None,
        filters: typing.List[DTOFieldFilter] = None,
        pk: bool = False,
        use_default_validator: bool = True,
        default_value: typing.Union[typing.Callable, typing.Any] = None,
        partition_data: bool = False,
        convert_to_entity: typing.Callable = None,
        convert_from_entity: typing.Callable = None,
        unique: str = None,
        candidate_key: bool = False,
        search: bool = True,
        read_only: bool = False,
    ):
        """
        -----------
        Parameters:
        -----------
        type: Tipo esperado para a propriedade. Se for do tipo enum.Enum, o valor recebido, para atribuição à propriedade, será convertido para o enumerado.
        not_null: O campo não poderá ser None, ou vazio, no caso de strings.
        resume: O campo será usado como resumo, isto é, será sempre rotornado num HTTP GET que liste os dados (mesmo que não seja solicitado por meio da query string "fields").
        min: Menor valor permitido (ou menor comprimento, para strings).
        max: Maior valor permitido (ou maior comprimento, para strings)
        validator: Função que recebe o valor (a ser atribuído), e retorna o mesmo valor após algum tipo de tratamento (como adição ou remoção, automática, de formatação).
        strip: O valor da string sofrerá strip (remoção de espaços no início e no fim), antes de ser guardado (só é útil para strings).
        entity_field: Nome da propriedade equivalente na classe de entity (que reflete a estruturua do banco de dados).
        filters: Lista de filtros adicionais suportados para esta propriedade (adicionais, porque todos as propriedades, por padrão, suportam filtros de igualdade, que podem ser passados por meio de uma query string, com mesmo nome da proriedade, e um valor qualquer a ser comparado).
          Essa lista de filtros consiste em objetos do DTOFieldFilter (veja a documentação da classe para enteder a estrutura de declaração dos filtros).
        pk: Flag indicando se o campo corresponde à chave da entidade corresponednte.
        use_default_validator: Flag indicando se o validator padrão deve ser aplicado à propriedade (esse validator padrão verifica o tipo de dados passado, e as demais verificações recebidas no filed, como, por exemplo, valor máximo, mínio, not_null, etc).
        default_value: Valor padrão de preenchimento da propriedade, caso não se receba conteúdo para a mesma (podendo ser um valor estático, ou uma função a ser chamada no preenchimento).
        partition_data: Flag indicando se esta propriedade participa dos campos de particionamento da entidade, isto é, campos sempre usados nas queries de listagem gravação dos dados, inclusíve para recuperação de entidades relacionadas.
        convert_to_entity: Função para converter o valor contido no DTO, para o(s) valor(es) a serem gravados no objeto de entidade (durante a conversão). É útil para casos onde não há equivalência um para um entre um campo do DTO e um da entidade
          (por exemplo, uma chave de cnpj que pode ser guardada em mais de um campo do BD). Outro caso de uso, é quando um campo tem formatação diferente entre o DTO e a entidade, carecendo de conversão customizada.
          A função recebida deve suportar os parâmetros (dto_value: Any, dto: DTOBase), e retornar um Dict[str, Any], como uma coleção de chaves e valores a serem atribuídos na entidade.
        convert_from_entity: Função para converter o valor contido na Entity, para o(s) valor(es) a serem gravados no objeto DTO (durante a conversão). É útil para casos onde não há equivalência um para um entre um campo do DTO e um da entidade
          (por exemplo, uma chave de cnpj que pode ser guardada em mais de um campo do BD). Outro caso de uso, é quando um campo tem formatação diferente entre o DTO e a entidade, carecendo de conversão customizada.
          A função recebida deve suportar os parâmetros (entity_value: Any, entity_fields: Dict[str, Any]), e retornar um Dict[str, Any], como uma coleção de chaves e valores a serem atribuídos no DTO.
        unique: Permite indicar um nome de chave de unicidade. Cada chave de unicidade é considerada no momento de uma inserção no BD (impedindo duplicações indesejadas).
        candidate_key: Permite indicar que este campo se trata de uma chave candidata (útil para operações unitárias, como GTE e DELETE, pois estas irão verificar se o tipo do dado recebido bate com a PK, ou com as chaves candidatas, para resolver como fará a query).
        search: Indica que esse campo é passível de busca, por meio do argumento "search" passado num GET List, como query string (por hora, apenas pesquisas simples, por meio de operador like, estão implementadas).
        read_only: Permite declarar propriedades que estão disponíveis no GET (list ou unitário), mas que não poderão ser usadas para gravação (POST, PUT ou PATCH).
        """
        self.name = None
        self.expected_type = type
        self.not_null = not_null
        self.resume = resume
        self.min = min
        self.max = max
        self.validator = validator
        self.strip = strip
        self.entity_field = entity_field
        self.filters = filters
        self.pk = pk
        self.use_default_validator = use_default_validator
        self.default_value = default_value
        self.partition_data = partition_data
        self.convert_to_entity = convert_to_entity
        self.convert_from_entity = convert_from_entity
        self.unique = unique
        self.candidate_key = candidate_key
        self.search = search
        self.read_only = read_only

        self.storage_name = f"_{self.__class__.__name__}#{self.__class__._ref_counter}"
        self.__class__._ref_counter += 1

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return instance.__dict__[self.storage_name]

    def __set__(self, instance, value):
        try:
            if self.validator is None:
                if self.use_default_validator:
                    value = self.validate(self, value, instance)
            else:
                if self.use_default_validator:
                    value = self.validate(self, value, instance)
                value = self.validator(self, value)
        except ValueError as e:
            if not (
                "escape_validator" in instance.__dict__
                and instance.__dict__["escape_validator"] == True
            ):
                raise

        instance.__dict__[self.storage_name] = value

    def validate(self, dto_field: "DTOField", value, instance):
        """
        Default validator (ckecking default constraints: not null, type, min, max and enum types).
        """

        # Checking not null constraint
        if (
            self.not_null
            and (value is None or (isinstance(value, str) and len(value.strip()) <= 0))
            and (
                not dto_field.pk
                or (
                    "generate_default_pk_value" in instance.__dict__
                    and instance.__dict__["generate_default_pk_value"]
                )
            )
        ):
            raise ValueError(f"O campo {self.storage_name} deve estar preenchido.")

        # Checking type constraint
        # TODO Ver como suportar typing
        if (
            self.expected_type is not None
            and not isinstance(value, self.expected_type)
            and value is not None
        ):
            value = TypeValidatorUtil.validate(self, value)

        # Checking min constraint
        if self.min is not None:
            if isinstance(value, str) and (len(value) < self.min):
                raise ValueError(
                    f"O campo {self.storage_name} deve conter no mínimo {self.min} caracteres. Valor recebido: {value}."
                )
            elif (
                isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, Decimal)
            ) and (value < self.min):
                raise ValueError(
                    f"O campo {self.storage_name} deve ser maior ou igual a {self.min}. Valor recebido: {value}."
                )

        # Checking min constraint
        if self.max is not None:
            if isinstance(value, str) and (len(value) > self.max):
                raise ValueError(
                    f"O campo {self.storage_name} deve conter no máximo {self.max} caracteres. Valor recebido: {value}."
                )
            elif (
                isinstance(value, int)
                or isinstance(value, float)
                or isinstance(value, Decimal)
            ) and (value > self.max):
                raise ValueError(
                    f"O campo {self.storage_name} deve ser menor ou igual a {self.max}. Valor recebido: {value}."
                )

        # Striping strings (if desired)
        if isinstance(value, str) and self.strip:
            value = value.strip()

        return value

    def get_entity_field_name(self) -> str:
        """
        Retorna o nome correspondente do field no entity
        (o qual é o nome do field no DTO por padrão, ou o nome que for
        passado no parâmetro "entity_field" no construtor).
        """

        if self.entity_field is not None:
            return self.entity_field
        else:
            return self.name
